#! /usr/bin/env python

import os, sys

""" Modifies "GnuCEmitter.g.fixed" so both the parse tree and the tokens are printed.

               ./insert_prints.py <rule> <token>

        where <rule> and <token> are "ALL", "SOME", or "NONE". The
        output is copied to the 'grammars" directory and a "make" is
        done there.

The user can control which parts of the parse tree are printed.

1. Modifications to "GnuCEmitter.g"

GnuCEmitter.g needs to be regularized slightly:
    Remove spaces at the end of lines.
    Rules should be separated by a single blank line.
    Remove blank lines from inside rules.
    Remove the four spaces in front of "unaryOperator".
    Replace each action of the form
        print( variable );
    by
        <@ variable @>
    Replace tabs with 8 spaces.
    Remove commented out exception handling.
    Move the colon in "attributeDecl:" to the next line.
    Replace line 168
        currentOutput.print( t.getText() + " " );
    with
        currentOutput.print( "\"" + t.getText() + "\" " );

I have made changes in the "top" part of "GnuCEmitter.g.fixed". The print
function was changed to put quotes around all the tokens. ";", ",", "(", and
")" are printed as SEMI, COMMA, LPARENS, and RPARENS respectively. Maybe
other changes. Run a "diff" to find them.

2. Introduction

There are two ways in which this program can insert print statements. Here is
a rule from GnuCEmitter.g:

1 typeQualifier
2         :       a:"const"                       { print( a ); }
3         |       b:"volatile"                    { print( b ); }
4         ;

Here is the same rule in "GnuCEmitter.g.fixed":

1 typeQualifier
2         :       a:"const"                       { <@ a @> }
3         |       b:"volatile"                    { <@ b @> }
4         ;

"print"s for this rule can be put in three places:

1 typeQualifier
2   { System.out.print( "(" ); print( "typeQualifier" ); }
3         :  (
4                 a:"const"                       {  currentOutput.print( "(\"typeQualifier.0\", " ); print( a );  currentOutput.print( "), " ); }
5         |       b:"volatile"                    {  currentOutput.print( "(\"typeQualifier.1\", " ); print( b );  currentOutput.print( "), " ); }
6            )
7   { System.out.print( "), " ); }
8        ;

Line 2 and 7 cause some stuff to be printed whenever the rule is entered or
left. This is a "rule print". Lines 4 and 5 cause other stuff to be printed
whenever the token is printed. These are "token print"s.

The parentheses in lines 3 and 6 are added to ensure that line 7 is
executed. The parentheses create a subrule.

The general behavior of the program is set by the variables
    do_rules         do_tokens
    rule_codechar    token_codechar
    do_ats

3. Rule Prints

By default, rule_codechar is '@' and token_codechar is '#'. The line in each
rule containing the name of the rule (usually the first line) may start with
any combination of "@" and '#'. Any line of the rule containing a print
statement
      { ... print( something ); ... }
may start with a "#".

If do_tokens (do_rules) is:
    ALL   Process every line (rule) irrespective of any codechars.
    SOME  Decide whether to process a line (rule) by looking for codechars
          at the beginning of the line (rule).
    NONE  Do not process any line (rule). But a rule print is always added
          to the rule "externalList".

Token prints (rule prints) are added when the line starts with a '#' ('@').

The rule print for "externalList" is always added because we want
parentheses around the entire output.

4. Token Prints

A token_codechar in front of the rule name says to add ALL possible token
prints to the rule. Any token_codechars in front of lines of the rule are
ignored. See the function "SetRuleBehavior".

If there is no token_codechar in front of the rule name, put token prints in
only those lines starting with '#'.

5. Do_ats

If do_ats == True, "@@@" is printed between the declarations in the input
header file and no rule print is inserted for "externalDef".
"""
#>==================================================================

NONE, SOME, ALL = 0, 1, 2

def ClipLine(codechars, line):
    i = 0
    while line[i] in codechars:
        i += 1
    return line[i:]

# This class brings together the code that processes the text for a rule.
class RulePrints:
    templateR1 = '  { System.out.print( "(" ); ' \
          'print( "%s%s" ); }'
    templateR2 = '  { System.out.print( "(" ); ' \
          'print( "%s%s" ); %s'
    templateR3 = '   )\n  { ' \
           'System.out.print( "), " ); }\n        ;'
    def __init__(self, do_rules, rule_codechar, rule_codechar_bad,
                  token_codechar):
        self.do_rules = do_rules
        self.rule_codechar = rule_codechar
        self.rule_codechar_bad = rule_codechar_bad
        self.token_codechar = token_codechar
        self.codechars = (self.rule_codechar, rule_codechar_bad, self.token_codechar)

    # We want to convert everything after the ":" in the rule to a single
    # subrule. Then we can put a { ... } after the subrule and it will be
    # executed.
    # Note: This is about the only time I have ever used slice assignments
    # (not counting Numarray).
    def AddParens(self, line):
        i = 0
        while line[i] in self.codechars:
            i += 1
        j = i
        while line[j] == ' ':
            j += 1
        line1 = line[i:j+1] + '  ('
        line2 = line[:i] + '         ' + line[j+1:]
        return [line1, line2]

    def Rule_CodeChars(self, line):
        out = []
        i = 0
        while line[i] in self.codechars:
            if line[i] in (self.rule_codechar, self.rule_codechar_bad):
                out.append(line[i])
            i += 1
        return out
    
    def DoThisRule(self, firstline):
        warn = ' WARNING'
        ccs = self.Rule_CodeChars(firstline)
        if self.do_rules == ALL:
            if self.rule_codechar_bad in ccs:
                return True, warn
            return True, ''
        if self.do_rules == NONE:
            return False, ''
        if self.rule_codechar_bad in ccs:
            return True, warn
        if self.rule_codechar in ccs:
            return True, ''
        return False, ''

    def ProcessRule(self, lines):
        if lines[0] == 'protected':
            ruleline = 1
        else:
            ruleline = 0
        rule_name = ClipLine(self.codechars, lines[ruleline])

        out = []
        # A few rules start with "protected" instead of the rule name.
        if lines[0] == 'protected':
            out.append(lines[0])
            out.append(rule_name)
            doline, warn = self.DoThisRule(lines[ruleline])
            if doline:
                out.append(RulePrints.templateR1 % (rule_name, warn))
                lines[2:3] = self.AddParens(lines[2])
            lo = 2
        elif rule_name == 'initDecl':
            # {...}{...} is not allowed. This problem only occurs for "initDeclk".
            out.append(rule_name)
            doline, warn = self.DoThisRule(lines[ruleline])
            if doline:
                line1 = lines[1].strip()[2:]
                out.append(RulePrints.templateR2 % (rule_name, warn, line1))
                lines[2:3] = self.AddParens(lines[2])
            else:
                out.append(lines[1])
            lo = 2
        elif rule_name == 'externalList':
            # Always do this one to always surround the output with "()".
            out.append(rule_name)
            doline = True
            out.append(RulePrints.templateR1 % (rule_name, ''))
            lines[1:2] = self.AddParens(lines[1])
            lo = 1
        else:
            out.append(rule_name)
            doline, warn = self.DoThisRule(lines[0])
            if doline:
                out.append(RulePrints.templateR1 % (rule_name, warn))
                lines[1:2] = self.AddParens(lines[1])
                didthis = True
            lo = 1
        return rule_name, ruleline, lo, out, doline

class TokenPrints:
    templateT = ' currentOutput.print("(\\\"%s\\\", "); ' \
          'print(%s); ' \
          'currentOutput.print("), "); '

    def __init__(self, do_tokens, rule_codechar, rule_codechar_bad,
               token_codechar):
        self.rule_codechar = rule_codechar
        self.rule_codechar_bad = rule_codechar_bad
        self.token_codechar = token_codechar
        self.do_tokens = do_tokens
        self.codechars = (rule_codechar, rule_codechar_bad, token_codechar)

    def Token_CodeChar_Present(self, line):
        i = 0
        while line[i] in self.codechars:
            if line[i] == self.token_codechar:
                return True
            i += 1
        return False

    def SetRuleBehavior(self, firstline):
        if self.do_tokens == ALL:
            return ALL
        if self.do_tokens == NONE:
            return NONE
        # So now do_tokens must be SOME.
        if not self.Token_CodeChar_Present(firstline):
            return SOME
        return ALL

    def DoThisLine(self, rule_behavior, line):
        if rule_behavior == NONE:
            return False
        elif rule_behavior == ALL:
            return True
        else:
            return self.Token_CodeChar_Present(line)

    def ProcessLine(self, line, rule_name, count, rule_behavior):
        # Each <@ ... @> pair is found and processed.
        do_thisline = self.DoThisLine(rule_behavior, line)

        line = ClipLine(self.codechars, line)
        if not do_thisline:
            # Put default prints in.
            count += line.count('<@')
            line = line.replace('<@', 'print(')
            line = line.replace('@>', ');')
            return count, line

        # Process the line.
        end = 0
        start = line.find('<@', end)
        if start == -1:
            return count, line
        parts = []
        while 1:
            parts.append(line[end:start])
            end = line.find('@>', start)
            if end == -1:
                print >> sys.stderr, 'missing "@>" at'
                print >> sys.stderr, line
                sys.exit(0)
            end += 2  # Pass the "@>".
            var = line[start+2:end-2]
            newname = '%s.%i' % (rule_name, count)
            count += 1
            newprint = TokenPrints.templateT % (newname, var)
            parts.append(newprint)
            start = line.find('<@', end)
            if start == -1:
                break
        parts.append(line[end:])
        newline = ''.join(parts)
        return count, newline

def DoRule(text, ruleprints, tokenprints):
    rules = text.split('\n\n')
    out = []
    for rule in rules:
        out2 = []
        lines = rule.splitlines()
        try:
            rule_name, ruleline, lo, out2, didthis = \
                  ruleprints.ProcessRule(lines)
        except:
            print 'Error processing rule:'
            print lines
            raise
        rule_behavior = tokenprints.SetRuleBehavior(lines[ruleline])
    
        count = 0
        for line in lines[lo:]:
            count, lineout = tokenprints.ProcessLine(line, rule_name,
                        count, rule_behavior)
            out2.append(lineout)
        text = '\n'.join(out2)
        if didthis:
            text = text[:-1] + RulePrints.templateR3
        out.append(text)
    out.append('')
    return '\n'.join(out)    

def SplitFile(text):
    # Grammar files start with a bunch of java followed by the grammar
    # itself.  Separate the two.
    a = text.find('\nexternalList') + 1
    b = text.find('\n@externalList') + 1
    c = text.find('\n#externalList') + 1
    d = text.find('\n@#externalList') + 1
    e = text.find('\n#@externalList') + 1
    splitpoint = max(a, b, c, d, e)
    top = text[:splitpoint]
    # In Java, filename == class name.
    top = top.replace('GnuCEmitter', 'GnuCEmitter_print')
    bottom = text[splitpoint:].strip()
    return top, bottom

# Main program

if len(sys.argv) != 3:
    print >> sys.stderr, './insert_prints.py <rule> <token>'
    print >> sys.stderr, 'where <rule> and <token> are "ALL", "SOME", or "NONE".'
    sys.exit(0)
do_rules = sys.argv[1]
do_tokens = sys.argv[2]

if sys.argv[1] == 'ALL':
   do_rules = ALL
elif sys.argv[1] == 'SOME':
   do_rules = SOME
elif sys.argv[1] == 'NONE':
   do_rules = NONE
else:
   print >> sys.stderr, 'arguments must be "ALL", "SOME", or "NONE".'
   sys.exit(0)

if sys.argv[2] == 'ALL':
   do_tokens = ALL
elif sys.argv[2] == 'SOME':
   do_tokens = SOME
elif sys.argv[2] == 'NONE':
   do_tokens = NONE
else:
   print >> sys.stderr, 'arguments must be "ALL", "SOME", or "NONE".'
   sys.exit(0)

input_file = 'GnuCEmitter.g.fixed'
output_file = 'GnuCEmitter_print.g'
copied_file = '../../grammars/GnuCEmitter_print.g'
make_dir = '../../grammars'

token_codechar = '#'
rule_codechar = '@'
rule_codechar_bad = '%'

text = file(input_file, 'r').read()
top, bottom = SplitFile(text)

ruleprints = RulePrints(do_rules, rule_codechar, rule_codechar_bad,
        token_codechar)
tokenprints = TokenPrints(do_tokens, rule_codechar, rule_codechar_bad,
             token_codechar)

bottomout = DoRule(bottom, ruleprints, tokenprints)

f = file(output_file, 'w')
f.write(top)
f.write('\n\n')
f.write(bottomout)
f.close()
f = file(copied_file, 'w')
f.write(top)
f.write('\n\n')
f.write(bottomout)
f.close()
command = 'cd %s; make' % make_dir
os.system(command)
